from .constants.types import ErrorTypeStrings as ETS
from abc import ABC

from .utils.classes import find_non_abc_subclasses


class OQSBaseError(Exception, ABC):
    def __init__(self, message: str = "An error occurred while evaluating your expression!") -> None:
        super().__init__(message)
        self.readable_name: str = self.READABLE_NAME if hasattr(self, "READABLE_NAME") else ETS.BASE
        self.error_hierarchy: list[str] = self._build_error_hierarchy()

    def _build_error_hierarchy(self) -> list[str]:
        hierarchy: list[str] = []
        for cls in self.__class__.__mro__:
            if issubclass(cls, OQSBaseError) and hasattr(cls, 'READABLE_NAME'):
                hierarchy.append(cls.READABLE_NAME)
        hierarchy.append(ETS.BASE)
        return hierarchy


class OQSInvalidArgumentQuantityError(OQSBaseError):
    READABLE_NAME: str = ETS.INVALID_ARGUMENT_QUANTITY

    def __init__(self, function_name: str, expected_min: int, expected_max: int, actual: int) -> None:
        message: str = (
            f"Function '{function_name}' expected at least {expected_min} with a max of {expected_max} arguments, "
            f"but got {actual}"
        )
        super().__init__(message=message)


class OQSSyntaxError(OQSBaseError):
    READABLE_NAME: str = ETS.SYNTAX

    def __init__(self, message: str) -> None:
        super().__init__(message=message)


class OQSTypeError(OQSBaseError):
    READABLE_NAME: str = ETS.TYPE

    def __init__(self, message: str) -> None:
        super().__init__(message=message)


class OQSUndefinedVariableError(OQSBaseError):
    READABLE_NAME: str = ETS.UNDEFINED_VARIABLE

    def __init__(self, variable_name: str) -> None:
        message: str = f"The variable {variable_name} is not defined."
        super().__init__(message=message)


class OQSUndefinedFunctionError(OQSBaseError):
    READABLE_NAME: str = ETS.UNDEFINED_FUNCTION

    def __init__(self, function_name: str) -> None:
        message: str = f"The function '{function_name}' is not a valid function."
        super().__init__(message=message)


class OQSFunctionEvaluationError(OQSBaseError):
    READABLE_NAME: str = ETS.FUNCTION_EVALUATION

    def __init__(self, function_name: str, message: str) -> None:
        full_message: str = f"Error in function '{function_name}': {message}"
        super().__init__(message=full_message)


class OQSDivisionByZeroError(OQSBaseError):
    READABLE_NAME: str = ETS.DIVISION_BY_ZERO

    def __init__(self) -> None:
        super().__init__(message="Division by zero results in undefined.")


class OQSUnexpectedCharacterError(OQSSyntaxError):
    READABLE_NAME: str = ETS.UNEXPECTED_CHARACTER

    def __init__(self, message: str) -> None:
        super().__init__(message=message)


class OQSMissingExpectedCharacterError(OQSSyntaxError):
    READABLE_NAME: str = ETS.MISSING_EXPECTED_CHARACTER

    def __init__(self, message: str) -> None:
        super().__init__(message=message)


class OQSCustomErrorParent(OQSBaseError, ABC):
    READABLE_NAME: str = ETS.CUSTOM

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


def get_error_name_mapping() -> dict[str, type[OQSBaseError]]:
    error_classes: list[type[OQSBaseError]] = find_non_abc_subclasses(OQSBaseError)
    error_name_mapping: dict[str, type[OQSBaseError]] = {}
    for cls in error_classes:
        if hasattr(cls, "READABLE_NAME"):
            error_name_mapping[cls.READABLE_NAME.upper()] = cls
        else:
            raise TypeError(f"Class {cls.__name__} does not have a 'READABLE_NAME' class attribute.")
    return error_name_mapping



import unittest
from python_oqs_implementation.oqs.constants.types import ErrorTypeStrings as ETS
from .utils import language_engine_expected_result


class TestLanguageEngine(unittest.TestCase):
    def setUp(self):
        self.cases: dict[str, list[dict[str, any]]] = {}

    def leer(self, *args, **kwargs) -> None:
        language_engine_expected_result(self, *args, **kwargs)

    def test_multiply(self):
        self.leer("2 * 5", 10)

    def test_divide(self):
        self.leer("10 / 2", 5)

    def test_modulo(self):
        self.leer("9 % 2", 1)

    def test_logical_and_operator(self):
        self.leer("true & false", False)
        self.leer("1 & 0", False)
        self.leer('"text" & ""', False)

    def test_logical_or_operator(self):
        self.leer("false | true", True)
        self.leer("0 | 1", True)
        self.leer('"" | "text"', True)

    def test_string_concatenation(self):
        self.leer('"Hello " + "World"', "Hello World")

    def test_string_repetition(self):
        self.leer('"repeat" * 2', "repeatrepeat")

    def test_boolean_comparison(self):
        self.leer("true == false", False)
        self.leer("true != false", True)

    def test_list_concatenation(self):
        self.leer("[1, 2] + [3, 4]", [1, 2, 3, 4])

    def test_list_subtraction(self):
        self.leer("[1, 2, 3, 4] - [3]", [1, 2, 4])

    def test_kvs_concatenation(self):
        self.leer('{ "a": 1, "b": 2 } + { "c": 3 }', {"a": 1, "b": 2, "c": 3})

    def test_add_function(self):
        self.leer("ADD(1, 2)", 3)

    def test_complex_expression(self):
        self.leer('IF(LEN("test") == 4, "valid", "invalid")', "valid")
        self.leer('ADd(length([5, 10, ***[4, 3, "HEllo:,"]]), 20)', 25)

    def test_addition(self):
        self.leer("1 + 2", 3)

    def test_subtraction(self):
        self.leer("5 - 2", 3)

    def test_exponentiation(self):
        self.leer("2 ** 3", 8)

    def test_string_subtraction(self):
        self.leer('"remove" - "move"', "re")

    def test_list_add_function(self):
        self.leer('ADD([1, 2], [3, 4])', [1, 2, 3, 4])

    def test_kvs_add_function(self):
        self.leer('ADD({ "a": 1 }, { "b": 2 })', {"a": 1, "b": 2})

    def test_integer_function(self):
        self.leer('INTEGER(3.5)', 3)

    def test_decimal_function(self):
        self.leer('DECIMAL("42")', 42.0)

    def test_string_function(self):
        self.leer('STRING([1, 2, 3])', "[1, 2, 3]")

    def test_boolean_function(self):
        self.leer('BOOLEAN(1)', True)

    def test_keys_function(self):
        self.leer('KEYS({ "name": "OQS", "type": "script" })', ["name", "type"])

    def test_values_function(self):
        self.leer('VALUES({ "name": "OQS", "type": "script" })', ["OQS", "script"])

    def test_complex_if_expression(self):
        self.leer('IF(LEN("test") == 4, "valid", "invalid")', "valid")

    def test_complex_add_expression(self):
        self.leer('ADD(***[1, 2, 3, 4])', 10)

    def test_unpacked_integer(self):
        self.leer('INTEGER(***["5"])', 5)

    def test_kvs_expansion(self):
        self.leer('{***{"key1": "value1"}, ***{"key2": "value2"}}', {"key1": "value1", "key2": "value2"})

    def test_string_embedded_expression(self):
        self.leer('<{3 + 5}> is the answer', "8 is the answer", string_embedded=True)


class TestLanguageEngineAdvanced(unittest.TestCase):
    def setUp(self):
        self.cases: dict[str, list[dict[str, any]]] = {}

    def leer(self, *args, **kwargs) -> None:
        language_engine_expected_result(self, *args, **kwargs)

    def test_basic(self):
        self.leer("2 * 5 + 3", 13)
        self.leer('"Hello " + name', "Hello World", variables={"name": "World"})
        self.leer('ADD(1, 2, 3)', 6)
        self.leer('SUBTRACT(10, LENGTH(name))', 7, variables={"name": "OQS"})
        self.leer("MULTIPLY(3, ***numbers)", 12, variables={"numbers": [2, 2]})
        self.leer('DECIMAL(STRING(number))', 123.0, variables={"number": 123})
        self.leer('IF(LEN(text) > 5, "Long", "Short")', "Short", variables={"text": "Hello"})
        self.leer('{***kvs1, ***kvs2}', {"a": 1, "b": 2}, variables={"kvs1": {"a": 1}, "kvs2": {"b": 2}})
        self.leer('SUM([1, 2, 3]) + SUM([4, 5, 6])', 21)
        self.leer('STRING(ADD(1, 2, 3)) + " is the sum"', "6 is the sum")

    def test_numerical_and_string_operations(self):
        self.leer('3 ** 2 + LEN("test")', 13)
        self.leer('STRING(5.0 * 2) - ".0"', "10")

    def test_list_and_kvs_operations(self):
        self.leer('[1, 2, 3] + REVERSE([4, 5, 6])', [1, 2, 3, 6, 5, 4])
        self.leer('KEYS({ "a": 1, "b": 2 }) + VALUES({ "a": 1, "b": 2 })', ["a", "b", 1, 2])
        self.leer('SUM([1, 2, 3, 4])', 10)
        self.leer('MAX(***numbers)', 5, variables={"numbers": [1, 2, 3, 4, 5]})

    def test_complex_nested_expressions(self):
        self.leer('IF(LEN(name) > 5, SUBTRACT(LEN(name), 3), ADD(LEN(name), 2))', 6, variables={"name": "Scripting"})
        self.leer('STRING(ADD(***[INTEGER("1"), DECIMAL("2.5"), 3]))', "6.5")

    def test_error_handling(self):
        self.leer(
            'DIVIDE(10, 0)',
            expected_type=ETS.DIVISION_BY_ZERO,
            expect_error=True,
            error_message='Division by zero results in undefined.'
        )

    def test_string_embedded_expressions(self):
        self.leer('<{ADD(1, 2)}> + <{MULTIPLY(3, 4)}>', "3 + 12", string_embedded=True)
        self.leer(
            'Result is <{IF(LEN(text) == 5, "five", "not five")}>.',
            'Result is five.',
            variables={"text": "Hello"},
            string_embedded=True
        )

    def test_advanced_scenarios(self):
        self.leer('UNIQUE([***numbers, ***REVERSE(numbers)])', [1, 2, 3], variables={"numbers": [1, 2, 3]})
        self.leer('REMOVE_ITEM([1, 2, 3, 2, 3, 3], 3, 2)', [1, 2, 2, 3])
        self.leer(
            'UPDATE(kvs, "newKey", IF(ACCESS(kvs, "key") == 5, 10, 0))',
            {"key": 5, "newKey": 10},
            variables={"kvs": {"key": 5}}
        )

    def test_advanced_numerical_operations(self):
        self.leer('MODULO(15, 4) + DIVIDE(20, 5)', 7)
        self.leer('EXPONENTIATE(ADD(2, 3), SUBTRACT(5, 2))', 125)

    def test_complex_string_manipulation(self):
        self.leer('STRING(LEN("Hello")) + " characters"', "5 characters")
        self.leer('"Start-" + MULTIPLY("A", 3) + "-End"', "Start-AAA-End")

    def test_list_and_kvs_combinations(self):
        self.leer('APPEND(UNIQUE([1, 2, 2, 3]), MAX(***[4, 5, 6]))', [1, 2, 3, 6])
        self.leer('ACCESS({***{"key": "value"}, ***{"another": SUM([1, 2, 3])}}, "another")', 6)

    def test_function_calls_with_complex_arguments(self):
        self.leer('STRING(ADD(INTEGER("3"), DECIMAL("2.5"), MULTIPLY(2, 2)))', "9.5")
        self.leer('BOOLEAN(LEN("test") == 4)', True)

    def test_nested_and_conditional_expressions(self):
        self.leer('IF(LEN(name) > 5, LEN(name), "short")', 15, variables={"name": "OpenQuickScript"})
        self.leer('STRING(ADD(1, IF(true, 2, 3)))', "3")

    def test_error_scenarios_and_edge_cases(self):
        self.leer(
            'DIVIDE("10", 2)',
            expected_type=ETS.TYPE,
            expect_error=True,
            error_message="Cannot divide type 'String' by type 'Integer'."
        )

    def test_string_embedded_and_unpacking_expressions(self):
        self.leer(
            "<{ADD(1, LEN(text))}> is the length", "6 is the length", variables={"text": "Hello"}, string_embedded=True
        )
        self.leer(
            '{ "sum": "<{SUM([***numbers])}>", "max": "<{MAX(***numbers)}>" }',
            '{ "sum": "10", "max": "4" }',
            variables={"numbers": [1, 2, 3, 4]},
            string_embedded=True
        )

    def test_complex_function_calls_and_list_manipulations(self):
        self.leer('REMOVE_ITEM([1, 2, 3, 4, 5], ADD(2, 1))', [1, 2, 4, 5])
        self.leer('APPEND(numbers, MULTIPLY(LENGTH(numbers), 2))', [1, 2, 3, 6], variables={"numbers": [1, 2, 3]})

    def test_kvs_and_list_integration(self):
        self.leer(
            'KEYS({ "name": "OQS", "version": "0.1" }) + VALUES({ "name": "OQS", "version": "0.1" })',
            ["name", "version", "OQS", "0.1"]
        )
        self.leer(
            'UPDATE(kvs, "new", ADD(ACCESS(kvs, "existing"), 10))',
            {"existing": 5, "new": 15},
            variables={"kvs": {"existing": 5}}
        )

    def test_advanced_mathematical_operations(self):
        self.leer('MODULO(ADD(15, 5), DIVIDE(20, 2))', 0)
        self.leer('EXPONENTIATE(SUBTRACT(10, 2), 3)', 512)

    def test_string_manipulation_and_comparisons(self):
        self.leer('LEN("Concatenate") == 11', True)
        self.leer('STRING(LEN("Concatenate") == 11)', "true")
        self.leer('IF("text" - "t" == "ex", "Correct", "Incorrect")', "Correct")

    def test_complex_list_operations(self):
        self.leer('REVERSE(APPEND(numbers, SUM(numbers)))', [6, 3, 2, 1], variables={"numbers": [1, 2, 3]})

    def test_function_usage_with_nested_calls(self):
        self.leer('MULTIPLY(DECIMAL("3.5"), INTEGER(2.5))', 7.0)
        self.leer('ADD(STRING(LEN("abc")), STRING(LEN("def")))', "33")

    def test_conditional_expressions_with_functions(self):
        self.leer('IF(LEN(name) < 5, LEN(name), SUBTRACT(LEN(name), 2))', 4, variables={"name": "Script"})
        self.leer('IF(true, SUM([1,2,3]), MULTIPLY(2,3))', 6)

    def test_error_handling_scenarios(self):
        self.leer(
            'ADD("5", 10)', expected_type=ETS.TYPE, expect_error=True, error_message="Cannot add 'String' and 'Integer'"
        )
        self.leer(
            'SUBTRACT([1, 2, 3], 1)',
            expected_type=ETS.TYPE,
            expect_error=True,
            error_message="Cannot subtract 'List' by 'Integer'"
        )

    def test_string_embedded_and_complex_unpacking(self):
        self.leer(
            "<{MULTIPLY(2, LEN(text))}> is twice the length",
            "8 is twice the length",
            variables={"text": "Test"},
            string_embedded=True
        )
        self.leer('SUM([***numbers, ***REVERSE(numbers)])', 12, variables={"numbers": [1, 2, 3]})

    def test_advanced_function_calls_and_kvs_manipulations(self):
        self.leer("REMOVE_ITEM(UNIQUE(numbers), MIN(***numbers), 1)", [2, 3], variables={"numbers": [1, 2, 1, 3]})
        self.leer(
            'UPDATE(kvs, "total", SUM(VALUES(kvs)))',
            {"a": 1, "b": 2, "total": 3},
            variables={"kvs": {"a": 1, "b": 2}}
        )

    def test_complex_evaluation_involving_multiple_data_types(self):
        self.leer('STRING(ADD(***[1, 2, 3], LEN("[1, 2, 3]")))', "15")
        self.leer("SUM([1, 2, 3]) == STRING(6)", False)

    def test_order_of_operations(self):
        self.leer("1 + 2 * 3", 7)
        self.leer("1 * 2 + 3", 5)
        self.leer("(1 + 2) * 3", 9)
        self.leer("2 ** 3 * 4", 32)
        self.leer("2 * 3 ** 2", 18)
        self.leer("18 / 2 * 3", 27)
        self.leer("18 / (2 * 3)", 3)
        self.leer("4 + 3 - 2", 5)
        self.leer("4 - 3 + 2", 3)
        self.leer("4 * (2 + 3)", 20)
        self.leer("4 * 2 + 3", 11)
        self.leer("4 + 2 * 3", 10)
        self.leer("5 + 6 / 3 - 1", 6)
        self.leer("5 - 6 / 3 + 1", 4)


class TestOQSFunctions(unittest.TestCase):
    def setUp(self):
        self.cases: dict[str, list[dict[str, any]]] = {}

    def leer(self, *args, **kwargs) -> None:
        language_engine_expected_result(self, *args, **kwargs)

    def test_add_numbers(self):
        self.leer('ADD(1, 2)', 3)

    def test_add_strings(self):
        self.leer('ADD("Hello", " World")', "Hello World")

    def test_add_lists(self):
        self.leer('ADD([1, 2], [3, 4])', [1, 2, 3, 4])

    def test_add_kvs(self):
        self.leer('ADD({"a": 1}, {"b": 2})', {"a": 1, "b": 2})

    def test_subtract_numbers(self):
        self.leer('SUBTRACT(5, 3)', 2)

    def test_subtract_strings(self):
        self.leer('SUBTRACT("hello", "lo")', "hel")

    def test_subtract_lists(self):
        self.leer('SUBTRACT([1, 2, 3], [3])', [1, 2])

    def test_multiply_numbers(self):
        self.leer('MULTIPLY(3, 2)', 6)

    def test_multiply_string(self):
        self.leer('MULTIPLY("ab", 3)', "ababab")

    def test_divide_numbers(self):
        self.leer('DIVIDE(10, 2)', 5)

    def test_divide_by_zero(self):
        self.leer(
            'DIVIDE(5, 0)',
            expected_type=ETS.DIVISION_BY_ZERO,
            expect_error=True,
            error_message='Division by zero results in undefined.'
        )

    def test_exponentiate(self):
        self.leer('EXPONENTIATE(2, 3)', 8)

    def test_modulo(self):
        self.leer('MODULO(10, 3)', 1)

    def test_less_than(self):
        self.leer('LESS_THAN(1, 2, 3)', True)
        self.leer('LESS_THAN(1, 3, 2)', False)

    def test_greater_than(self):
        self.leer('GREATER_THAN(3, 2, 1)', True)
        self.leer('GREATER_THAN(3, 1, 2)', False)

    def test_less_than_or_equal(self):
        self.leer('LESS_THAN_OR_EQUAL(1, 2, 2)', True)
        self.leer('LESS_THAN_OR_EQUAL(2, 2, 1)', False)

    def test_greater_than_or_equal(self):
        self.leer('GREATER_THAN_OR_EQUAL(3, 2, 2)', True)
        self.leer('GREATER_THAN_OR_EQUAL(2, 2, 3)', False)

    def test_equals(self):
        self.leer('EQUALS(2, 2, 2)', True)
        self.leer('EQUALS(2, 2, 3)', False)

    def test_not_equals(self):
        self.leer('NOT_EQUALS(2, 3, 4)', True)
        self.leer('NOT_EQUALS(2, 2, 2)', False)

    def test_strictly_equals(self):
        self.leer('STRICTLY_EQUALS("2", "2", "2")', True)
        self.leer('STRICTLY_EQUALS("2", "2", 2)', False)

    def test_strictly_not_equals(self):
        self.leer('STRICTLY_NOT_EQUALS("2", "3", 2)', True)
        self.leer('STRICTLY_NOT_EQUALS("2", "2", "2")', False)

    def test_integer_conversion(self):
        self.leer('INTEGER(3.5)', 3)

    def test_decimal_conversion(self):
        self.leer('DECIMAL("42")', 42.0)

    def test_string_conversion(self):
        self.leer('STRING([1, 2, 3])', "[1, 2, 3]")

    def test_boolean_conversion(self):
        self.leer('BOOLEAN(1)', True)

    def test_and_function(self):
        self.leer('AND(true, 1, "text")', True)
        self.leer('AND(true, 0)', False)
        self.leer('AND(false, "non-empty string")', False)

    def test_or_function(self):
        self.leer('OR(false, 0, null, "text")', True)
        self.leer('OR(false, 0, "")', False)
        self.leer('OR(null, 0, false, "")', False)

    def test_keys_function(self):
        self.leer('KEYS({"name": "OQS", "type": "script"})', ["name", "type"])

    def test_values_function(self):
        self.leer('VALUES({"name": "OQS", "type": "script"})', ["OQS", "script"])

    def test_unique_function(self):
        self.leer('UNIQUE([1, 2, 2, 3])', [1, 2, 3])

    def test_reverse_function(self):
        self.leer('REVERSE([1, 2, 3])', [3, 2, 1])

    def test_max_function(self):
        self.leer('MAX(1, 3, 2)', 3)

    def test_min_function(self):
        self.leer('MIN(1, 3, 2)', 1)

    def test_sum_function(self):
        self.leer('SUM([1, 2, 3])', 6)

    def test_length_function(self):
        self.leer('LENGTH("Hello")', 5)

    def test_append_function(self):
        self.leer('APPEND([1, 2], 3)', [1, 2, 3])

    def test_update_function(self):
        self.leer('UPDATE({"a": 1}, "b", 2)', {"a": 1, "b": 2})

    def test_remove_item_function(self):
        self.leer('REMOVE_ITEM([1, 2, 3, 2, 3, 3], 3, 2)', [1, 2, 2, 3])

    def test_access_function(self):
        self.leer('ACCESS({"a": 1, "b": 2}, "b")', 2)

    def test_if_function(self):
        self.leer('IF(1 > 2, "Yes", "No")', "No")

    def test_type_function(self):
        self.leer('TYPE(5)', "Integer")

    def test_is_type_function(self):
        self.leer('IS_TYPE(5, "number")', True)

    def test_try_function(self):
        self.leer('TRY(1/0, "Division By Zero Error", "Infinity")', "Infinity")

    def test_range_function(self):
        self.leer('RANGE(1, 4)', [1, 2, 3])

    def test_for_map_function(self):
        self.leer('FOR([1, 2, 3], "item", item * 2)', [2, 4, 6])

    def test_raise_function(self):
        self.leer(
            'RAISE("Custom Error", "An error occurred")',
            expected_type=ETS.CUSTOM,
            expect_error=True,
            error_message="An error occurred"
        )

    def test_filter_list(self):
        self.leer('FILTER([1, 2, 3, 4], "x", x > 2)', [3, 4])

    def test_filter_kvs(self):
        self.leer('FILTER({"a": 1, "b": 2, "c": 3}, "value", value == 2)', {"b": 2})

    # Tests for the SORT function
    def test_sort_list_ascending(self):
        self.leer('SORT([4, 1, 3, 2], "x", x)', [1, 2, 3, 4])

    def test_sort_list_descending(self):
        self.leer('SORT([4, 1, 3, 2], "x", x, true)', [4, 3, 2, 1])

    # Tests for the FLATTEN function
    def test_flatten_list(self):
        self.leer('FLATTEN([[1, 2], [3, 4], [5]])', [1, 2, 3, 4, 5])

    # Tests for the SLICE function
    def test_slice_list(self):
        self.leer('SLICE([1, 2, 3, 4, 5], 1, 3)', [2, 3])

    def test_slice_string(self):
        self.leer('SLICE("Hello World", 0, 5)', "Hello")

    # Tests for the IN function
    def test_in_list(self):
        self.leer('IN(3, [1, 2, 3, 4])', True)

    def test_in_kvs(self):
        self.leer('IN("b", {"a": 1, "b": 2, "c": 3})', True)

    def test_not_in_list(self):
        self.leer('IN("z", [1, 2, 3, 4])', False)

    def test_not_in_kvs(self):
        self.leer('IN("d", {"a": 1, "b": 2, "c": 3})', False)



from typing import Callable
from . import built_in_functions
from .errors import (
    OQSUndefinedFunctionError,
    OQSBaseError,
    OQSFunctionEvaluationError,
    OQSUndefinedVariableError,
    OQSSyntaxError,
    OQSTypeError,
)
from .nodes import (
    FunctionNode,
    ASTNode,
    NumberNode,
    StringNode,
    NullNode,
    ListNode,
    VariableNode,
    BinaryOpNode,
    KVSNode,
    BooleanNode,
    UnparsedNode,
    ComparisonOpNode,
    PackedNode,
    EvaluatedNode
)
from .parser import OQSParser


class OQSInterpreter:
    OPERATORS: dict[str, str] = {
        '+': "ADD",
        '-': "SUBTRACT",
        '*': "MULTIPLY",
        '/': "DIVIDE",
        '%': "MODULO",
        '**': "EXPONENTIATE",
        '<': "LESS_THAN",
        '>': "GREATER_THAN",
        '<=': "LESS_THAN_OR_EQUAL",
        '>=': "GREATER_THAN_OR_EQUAL",
        '==': "EQUALS",
        '!=': "NOT_EQUALS",
        '===': "STRICTLY_EQUALS",
        '!==': "STRICTLY_NOT_EQUALS",
        '&': "AND",
        '|': "OR"
    }
    FUNCTIONS: dict[str, Callable] = {
        "ADD": built_in_functions.bif_add,
        "SUBTRACT": built_in_functions.bif_subtract,
        "MULTIPLY": built_in_functions.bif_multiply,
        "DIVIDE": built_in_functions.bif_divide,
        "EXPONENTIATE": built_in_functions.bif_exponentiate,
        "MODULO": built_in_functions.bif_modulo,
        "LESS_THAN": built_in_functions.bif_less_than,
        "GREATER_THAN": built_in_functions.bif_greater_than,
        "LESS_THAN_OR_EQUAL": built_in_functions.bif_less_than_or_equal,
        "GREATER_THAN_OR_EQUAL": built_in_functions.bif_greater_than_or_equal,
        "EQUALS": built_in_functions.bif_equals,
        "NOT_EQUALS": built_in_functions.bif_not_equals,
        "STRICTLY_EQUALS": built_in_functions.bif_strictly_equals,
        "STRICTLY_NOT_EQUALS": built_in_functions.bif_strictly_not_equals,
        "AND": built_in_functions.bif_and,
        "OR": built_in_functions.bif_or,
        "INTEGER": built_in_functions.bif_integer,
        "DECIMAL": built_in_functions.bif_decimal,
        "STRING": built_in_functions.bif_string,
        "LIST": built_in_functions.bif_list,
        "KVS": built_in_functions.bif_kvs,
        "BOOLEAN": built_in_functions.bif_boolean,
        "BOOL": built_in_functions.bif_boolean,
        "KEYS": built_in_functions.bif_keys,
        "VALUES": built_in_functions.bif_values,
        "UNIQUE": built_in_functions.bif_unique,
        "REVERSE": built_in_functions.bif_reverse,
        "MAX": built_in_functions.bif_max,
        "MIN": built_in_functions.bif_min,
        "SUM": built_in_functions.bif_sum,
        "LENGTH": built_in_functions.bif_length,
        "LEN": built_in_functions.bif_length,
        "APPEND": built_in_functions.bif_append,
        "UPDATE": built_in_functions.bif_update,
        "REMOVE_ITEM": built_in_functions.bif_remove_item,
        "REMOVE": built_in_functions.bif_remove,
        "ACCESS": built_in_functions.bif_access,
        "IF": built_in_functions.bif_if,
        "TYPE": built_in_functions.bif_type,
        "IS_TYPE": built_in_functions.bif_is_type,
        "TRY": built_in_functions.bif_try,
        "RANGE": built_in_functions.bif_range,
        "FOR": built_in_functions.bif_for_or_map,
        "MAP": built_in_functions.bif_for_or_map,
        "RAISE": built_in_functions.bif_raise,
        "FILTER": built_in_functions.bif_filter,
        "SORT": built_in_functions.bif_sort,
        "FLATTEN": built_in_functions.bif_flatten,
        "SLICE": built_in_functions.bif_slice,
        "IN": built_in_functions.bif_in
    }

    def __init__(self, expression: str, variables: dict[str, any] | None = None) -> None:
        self.original_expression: str = expression
        self.parser: OQSParser = OQSParser()
        self.original_ast: ASTNode = self.parser.parse(expression=self.original_expression)
        self.variables: dict[str, any] = variables if variables else {}

    def add_additional_function(self, function_name: str, function: Callable):
        self.FUNCTIONS[function_name.upper()] = function

    def results(self) -> any:
        return self.evaluate(self.original_ast)

    def parse_and_evaluate(self, *args, **kwargs) -> any:
        return self.evaluate(self.parser.parse(*args, **kwargs))

    def evaluate(self, node: ASTNode) -> any:
        if isinstance(node, EvaluatedNode):
            return node.value
        elif isinstance(node, NumberNode):
            return node.value
        elif isinstance(node, StringNode):
            return node.value
        elif isinstance(node, NullNode):
            return None
        elif isinstance(node, ListNode):
            elements: list[any] = []
            for elem in node.elements:
                if isinstance(elem, PackedNode):
                    evaluated_elem: any = self.parse_and_evaluate(elem.expression)
                    if isinstance(evaluated_elem, list):
                        elements.extend(evaluated_elem)
                    else:
                        raise OQSTypeError(message="Cannot unpack anything into a list construction other than a List.")
                else:
                    elements.append(self.evaluate(elem))
            return elements
        elif isinstance(node, VariableNode):
            if node.name in self.variables:
                return self.variables[node.name]
            else:
                raise OQSUndefinedVariableError(node.name)
        elif isinstance(node, BinaryOpNode):
            if node.op in self.OPERATORS:
                function_node: FunctionNode = FunctionNode(name=self.OPERATORS[node.op], args=[node.left, node.right])
                return self.FUNCTIONS[function_node.name](self, function_node)
            else:
                raise OQSSyntaxError(f"Invalid binary operator '{node.op}'")
        elif isinstance(node, ComparisonOpNode):
            if node.op in self.OPERATORS:
                function_node: FunctionNode = FunctionNode(name=self.OPERATORS[node.op], args=[node.left, node.right])
                return self.FUNCTIONS[function_node.name](self, function_node)
            else:
                raise OQSSyntaxError(f"Invalid comparison operator '{node.op}'")
        elif isinstance(node, FunctionNode):
            try:
                if node.name.upper() in self.FUNCTIONS:
                    args: list[ASTNode] = []
                    for arg in node.args:
                        if isinstance(arg, PackedNode):
                            parsed_value: any = self.parse_and_evaluate(arg.expression)
                            if isinstance(parsed_value, list):
                                args.extend([EvaluatedNode(part) for part in parsed_value])
                            else:
                                raise OQSTypeError(
                                    message="Cannot unpack anything into a function call other than a List."
                                )
                        else:
                            args.append(arg)
                    return self.FUNCTIONS[node.name.upper()](self, FunctionNode(name=node.name, args=args))
                else:
                    raise OQSUndefinedFunctionError(function_name=node.name)
            except OQSBaseError:
                raise
            except Exception as e:
                raise OQSFunctionEvaluationError(function_name=node.name, message=str(e))
        elif isinstance(node, KVSNode):
            kvs: dict[str, any] = {}
            for key, value in node.key_value_store.items():
                if isinstance(value, PackedNode):
                    unpacked_node: any = self.parse_and_evaluate(value.expression)
                    if isinstance(unpacked_node, list):
                        unpacked_kvs: dict[str, any] = self.evaluate(
                            FunctionNode(name="UNPACKED_KVS", args=unpacked_node)
                        )
                        for unpacked_key, unpacked_value in unpacked_kvs.items():
                            kvs[unpacked_key] = unpacked_value
                    elif isinstance(unpacked_node, dict):
                        for unpacked_key, unpacked_value in unpacked_node.items():
                            kvs[unpacked_key] = unpacked_value
                    else:
                        raise OQSTypeError(
                            message="Cannot unpack anything into a KVS construction other than a List or KVS."
                        )
                else:
                    kvs[self.evaluate(key)] = self.evaluate(value)
            return kvs
        elif isinstance(node, BooleanNode):
            return node.value
        elif isinstance(node, UnparsedNode):
            return self.parse_and_evaluate(node.token)
        else:
            raise OQSSyntaxError(f"Unable to parse the following: {node}")



import unittest
from typing import Callable
from python_oqs_implementation.oqs.errors import (
    OQSSyntaxError, OQSUnexpectedCharacterError, OQSMissingExpectedCharacterError
)
from python_oqs_implementation.oqs.parser import OQSParser


class TestSingleTokenizer(unittest.TestCase):
    def setUp(self) -> None:
        self._parser: OQSParser = OQSParser()
        self.tokenize: Callable = self._parser.tokenize_expression

    def test_string(self):
        self.assertEqual(['"string"'], self.tokenize('"string"'))
        self.assertEqual(["'string'"], self.tokenize("'string'"))

        with self.assertRaises(OQSUnexpectedCharacterError):
            self.tokenize('"string".')

        with self.assertRaises(OQSSyntaxError):
            self.tokenize('"string" "string"')

    def test_variable(self):
        self.assertEqual(['variable'], self.tokenize('variable'))
        self.assertEqual(['vari3ble'], self.tokenize('vari3ble'))
        self.assertEqual(['_variable'], self.tokenize('_variable'))
        self.assertEqual(['varia_ble'], self.tokenize('varia_ble'))
        self.assertEqual(['variable_'], self.tokenize('variable_'))

        with self.assertRaises(OQSUnexpectedCharacterError):
            self.tokenize('variable.')

        with self.assertRaises(OQSMissingExpectedCharacterError):
            self.tokenize('variable variable')

    def test_integer(self):
        self.assertEqual(['5'], self.tokenize('5'))
        self.assertEqual(['555'], self.tokenize('555'))
        self.assertEqual(['555555'], self.tokenize('555_555'))
        self.assertEqual(['5555555'], self.tokenize('5_555_555'))

        with self.assertRaises(OQSUnexpectedCharacterError):
            self.tokenize('5s5')

        with self.assertRaises(OQSSyntaxError):
            self.tokenize('5 5')

    def test_decimal(self):
        self.assertEqual(['5.5'], self.tokenize('5.5'))
        self.assertEqual(['5.555555'], self.tokenize('5.555555'))
        self.assertEqual(['0.5'], self.tokenize('.5'))
        self.assertEqual(['5.0'], self.tokenize('5.'))
        self.assertEqual(['555555.5'], self.tokenize('555_555.5'))

        with self.assertRaises(OQSUnexpectedCharacterError):
            self.tokenize('.')

        with self.assertRaises(OQSUnexpectedCharacterError):
            self.tokenize('55.5s')

        with self.assertRaises(OQSUnexpectedCharacterError):
            self.tokenize('555.5_5')

        with self.assertRaises(OQSSyntaxError):
            self.tokenize('555.55 5.55')

    def test_function(self):
        self.assertEqual(['ADD()'], self.tokenize('ADD()'))
        self.assertEqual(['ADD(variable)'], self.tokenize('ADD(variable)'))
        self.assertEqual(['ADD(variable, variable,variable)'], self.tokenize('ADD(variable, variable,variable)'))
        self.assertEqual(['ADD("string")'], self.tokenize('ADD("string")'))
        self.assertEqual(['ADD(555)'], self.tokenize('ADD(555)'))
        self.assertEqual(['ADD(5.5)'], self.tokenize('ADD(5.5)'))
        self.assertEqual(['ADD(5 5 5 5 )'], self.tokenize('ADD(5 5 5 5 )'))
        self.assertEqual(['ADD(5 + 5)'], self.tokenize('ADD(5 + 5)'))

        with self.assertRaises(OQSMissingExpectedCharacterError):
            self.tokenize('ADD(')

        with self.assertRaises(OQSMissingExpectedCharacterError):
            self.tokenize('ADD(5')

        with self.assertRaises(OQSUnexpectedCharacterError):
            self.tokenize('ADD(5).')

        with self.assertRaises(OQSSyntaxError):
            self.tokenize('ADD(5) 5')

    def test_nested_function(self):
        self.assertEqual(['ADD(ADD(5, ADD(5, 2, 5)), 2)'], self.tokenize('ADD(ADD(5, ADD(5, 2, 5)), 2)'))
        self.assertEqual(['ADD({{{{)'], self.tokenize('ADD({{{{)'))
        self.assertEqual(['ADD([[[[)'], self.tokenize('ADD([[[[)'))

        with self.assertRaises(OQSMissingExpectedCharacterError):
            self.tokenize('ADD((()')

    def test_list(self):
        self.assertEqual(
            ['[5, 5, "string"' + ", 'string', 4, ADD(3, 4), 2]"],
            self.tokenize('[5, 5, "string"' + ", 'string', 4, ADD(3, 4), 2]"),
        )
        self.assertEqual(['[]'], self.tokenize('[]'))

        with self.assertRaises(OQSMissingExpectedCharacterError):
            self.tokenize('[')

        with self.assertRaises(OQSMissingExpectedCharacterError):
            self.tokenize('[5')

        with self.assertRaises(OQSUnexpectedCharacterError):
            self.tokenize('[5].')

        with self.assertRaises(OQSSyntaxError):
            self.tokenize('[5] 5')

    def test_nested_list(self):
        self.assertEqual(['[5, [5, 2], [5, 2, [5, 2, [5, 2]]]]'], self.tokenize('[5, [5, 2], [5, 2, [5, 2, [5, 2]]]]'))
        self.assertEqual(['[((((]'], self.tokenize('[((((]'))
        self.assertEqual(['[{{{{]'], self.tokenize('[{{{{]'))

        with self.assertRaises(OQSMissingExpectedCharacterError):
            self.tokenize('[[[]')

    def test_kvs(self):
        self.assertEqual(['{anything}'], self.tokenize('{anything}'))

        with self.assertRaises(OQSMissingExpectedCharacterError):
            self.tokenize('{')

        with self.assertRaises(OQSMissingExpectedCharacterError):
            self.tokenize('{"string": variable')

        with self.assertRaises(OQSUnexpectedCharacterError):
            self.tokenize('{"string": variable}.')

        with self.assertRaises(OQSSyntaxError):
            self.tokenize('{"string": variable} 5')

    def test_nested_kvs(self):
        self.assertEqual(
            ['{"string": variable, "string2": {"string": variable}}'],
            self.tokenize('{"string": variable, "string2": {"string": variable}}')
        )
        self.assertEqual(['{[[[[}'], self.tokenize('{[[[[}'))
        self.assertEqual(['{((((}'], self.tokenize('{((((}'))

        with self.assertRaises(OQSMissingExpectedCharacterError):
            self.tokenize('{{}')

    def test_unpacking(self):
        self.assertEqual(['***variable'], self.tokenize('***variable'))
        self.assertEqual(['***{}'], self.tokenize('***{}'))
        self.assertEqual(['***{variable_1:variable_2}'], self.tokenize('***{variable_1:variable_2}'))
        self.assertEqual(['***[]'], self.tokenize('***[]'))
        self.assertEqual(['***[variable]'], self.tokenize('***[variable]'))
        self.assertEqual(['***()'], self.tokenize('***()'))
        self.assertEqual(['***(variable_1)'], self.tokenize('***(variable_1)'))

        with self.assertRaises(OQSUnexpectedCharacterError):
            self.tokenize('***"string"')

        with self.assertRaises(OQSUnexpectedCharacterError):
            self.tokenize("***'string'")

        with self.assertRaises(OQSUnexpectedCharacterError):
            self.tokenize('***1')

        with self.assertRaises(OQSUnexpectedCharacterError):
            self.tokenize('***1.0')

    def operator_test(self, operator: str):
        self.assertEqual(['5', operator,  '3'], self.tokenize(f'5 {operator} 3'))
        self.assertEqual(['"string"', operator,  '3'], self.tokenize(f'"string" {operator} 3'))
        self.assertEqual(['variable', operator, '3'], self.tokenize(f'variable {operator} 3'))
        self.assertEqual(['[]', operator,  '3'], self.tokenize(f'[] {operator} 3'))
        self.assertEqual([f'(5 {operator} 5)', operator,  '3'], self.tokenize(f'(5 {operator} 5) {operator} 3'))
        self.assertEqual(['anything', operator, 'anything'], self.tokenize(f'anything {operator} anything'))
        self.assertEqual(['anything', operator, 'anything'], self.tokenize(f'anything{operator}anything'))
        self.assertEqual(
            ['anything', operator, 'anything', operator, 'anything'],
            self.tokenize(f'anything {operator} anything {operator} anything')
        )
        self.assertEqual(['anything', operator*4, 'anything'], self.tokenize(f'anything {operator*4} anything'))

        with self.assertRaises(OQSMissingExpectedCharacterError):
            self.tokenize(f'5 {operator} ')

        with self.assertRaises(OQSMissingExpectedCharacterError):
            self.tokenize(f'5 {operator} 5 {operator}')

        with self.assertRaises(OQSMissingExpectedCharacterError):
            self.tokenize(f'{operator} 5')

    def test_addition(self):
        self.operator_test(operator="+")

    def test_subtraction(self):
        self.operator_test(operator="-")

    def test_division(self):
        self.operator_test(operator="/")

    def test_multiplication(self):
        self.operator_test(operator="*")

    def test_exponentiation(self):
        self.operator_test(operator="**")

    def test_modulation(self):
        self.operator_test(operator="%")

    def test_greater_than(self):
        self.operator_test(operator=">")

    def test_less_than(self):
        self.operator_test(operator="<")

    def test_greater_than_or_equal_to(self):
        self.operator_test(operator=">=")

    def test_less_than_or_equal_to(self):
        self.operator_test(operator="<=")

    def test_equal(self):
        self.operator_test(operator="==")

    def test_not_equal(self):
        self.operator_test(operator="!=")

    def test_strictly_equal(self):
        self.operator_test(operator="===")

    def test_strictly_not_equal(self):
        self.operator_test(operator="!==")

    def test_commas(self):
        with self.assertRaises(OQSUnexpectedCharacterError):
            self.tokenize('5, 5')

        with self.assertRaises(OQSUnexpectedCharacterError):
            self.tokenize('variable, variable')

        with self.assertRaises(OQSUnexpectedCharacterError):
            self.tokenize('anything, ')


class TestSeparatorTokenizer(unittest.TestCase):
    def setUp(self) -> None:
        self._parser: OQSParser = OQSParser()
        self.separate: Callable = self._parser.separate_arguments

    def test_no_commas(self):
        self.assertEqual(['5 + 5'], self.separate("5 + 5"))
        self.assertEqual(['5'], self.separate('5'))
        self.assertEqual(['variable'], self.separate('variable'))
        self.assertEqual(['"string"'], self.separate('"string"'))
        self.assertEqual(['ADD(5, 4, 3)'], self.separate('ADD(5, 4, 3)'))

    def test_single_commas(self):
        self.assertEqual(['5 + 5'], self.separate('5 + 5,'))
        self.assertEqual(['5'], self.separate('5,'))
        self.assertEqual(['variable'], self.separate('variable,'))
        self.assertEqual(['"string"'], self.separate('"string",'))
        self.assertEqual(['ADD(5, 4, 3)'], self.separate('ADD(5, 4, 3),'))

    def test_two_commas(self):
        self.assertEqual(['5 + 5', '5 + 5'], self.separate('5 + 5, 5 + 5'))
        self.assertEqual(['5', '5'], self.separate('5, 5'))
        self.assertEqual(['variable', 'variable'], self.separate('variable, variable'))
        self.assertEqual(['"string"', '"string"'], self.separate('"string", "string"'))
        self.assertEqual(['ADD(5, 4, 3)', 'ADD(5, 4, 3)'], self.separate('ADD(5, 4, 3), ADD(5, 4, 3)'))

    def test_multiple_commas(self):
        self.assertEqual(['5 + 5', '5 + 5', '5 + 5'], self.separate('5 + 5, 5 + 5, 5 + 5'))
        self.assertEqual(['5', '5', '5'], self.separate('5, 5, 5'))
        self.assertEqual(['variable', 'variable', 'variable'], self.separate('variable, variable, variable'))
        self.assertEqual(['"string"', '"string"', '"string"'], self.separate('"string", "string", "string"'))
        self.assertEqual(
            ['ADD(5, 4, 3)', 'ADD(5, 4, 3)', 'ADD(5, 4, 3)'], self.separate('ADD(5, 4, 3), ADD(5, 4, 3), ADD(5, 4, 3)')
        )

    def test_commas_no_space(self):
        self.assertEqual(['5 + 5', '5 + 5', '5 + 5'], self.separate('5 + 5,5 + 5,5 + 5'))
        self.assertEqual(['5', '5', '5'], self.separate('5,5,5'))
        self.assertEqual(['variable', 'variable', 'variable'], self.separate('variable,variable,variable'))
        self.assertEqual(['"string"', '"string"', '"string"'], self.separate('"string","string","string"'))
        self.assertEqual(
            ['ADD(5, 4, 3)', 'ADD(5, 4, 3)', 'ADD(5, 4, 3)'], self.separate('ADD(5, 4, 3),ADD(5, 4, 3),ADD(5, 4, 3)')
        )

    def test_trailing_commas(self):
        self.assertEqual(['5 + 5', '5 + 5', '5 + 5'], self.separate('5 + 5, 5 + 5, 5 + 5,'))
        self.assertEqual(['5', '5', '5'], self.separate('5, 5, 5,'))
        self.assertEqual(['variable', 'variable', 'variable'], self.separate('variable, variable, variable,'))
        self.assertEqual(['"string"', '"string"', '"string"'], self.separate('"string", "string", "string",'))
        self.assertEqual(
            ['ADD(5, 4, 3)', 'ADD(5, 4, 3)', 'ADD(5, 4, 3)'], self.separate('ADD(5, 4, 3), ADD(5, 4, 3), ADD(5, 4, 3),')
        )



from setuptools import (setup, find_packages)


setup(
    name='oqs',
    version='0.8.1',
    packages=find_packages(include=['oqs', 'oqs.*']),
    description=
    "OQS (Open Quick Script) is a Python library for interpreting versatile expressions, supporting basic to advanced "
    "operations, custom functions, and performance monitoring. It efficiently handles fundamental types and "
    "operations, interprets expressions using variables from dictionaries or JSON, and adheres to robust error "
    "handling standards. OQS enhances Python's expression evaluation capabilities, making it ideal for diverse "
    "applications.",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Infuzu',
    author_email='oqs@infuzu.com',
    url='https://github.com/Infuzu/OQS',
    license='Creative Commons Attribution 4.0 International License',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
    license_files=('../LICENSE.md',)
)
